from typing import Tuple, Optional, Dict

from tensorflow.keras.models import Model
from tensorflow.keras.layers import Layer

from leap_model_rebuilder.modelexpander import expand_model
from leap_model_rebuilder.utils import is_custom_layers, convert_subclassed_to_functional, should_replace_layers


def rebuild_model(model: Model, input_tensor_shape: Optional[Tuple[int, ...]] = None,
                  layer_replacement_dict: Optional[Dict[str, Layer]] = None) -> Model:
    """
        Rebuild keras models from Subclassed to Functional and Expands custom layers inside

        :param model: keras model to rebuild in Functional way
        :type model: tensorflow.keras.Model
        :param input_tensor_shape: shape of input tensor
        :type input_tensor_shape: Tuple[int, ...]gst
        :param layer_replacement_dict: dictionary to replace layers.
               keys are the layer names and the value the layer to replace with.
        :return: rebuilt Functional keras model
        :rtype: tensorflow.keras.Model
        """
    if layer_replacement_dict is None:
        layer_replacement_dict = {}
    if input_tensor_shape is not None:
        model = convert_subclassed_to_functional(model, input_tensor_shape)
    while is_custom_layers(model) or should_replace_layers(model, layer_replacement_dict):
        model = expand_model(model, layer_replacement_dict)

    return model



import logging
from typing import List, Dict

import tensorflow as tf
from tensorflow.keras import Model
from tensorflow.keras.layers import Layer, InputLayer, Input
from tensorflow.python.keras.engine.node import Node, KerasHistory

from leap_model_rebuilder.utils import is_custom_layer

_logger = logging.getLogger(__name__)

"""The ModelExpander class expands custom layers inside keras models.
    The Expand process is using the call method of custom layers to expand to the logic inside them.
"""


def expand_model(model: Model, layer_replacement_dict: Dict[str, Layer]) -> Model:
    """
    Expanding model's custom layers

    :return: Expanded model
    :rtype: tensorflow.keras.Model
    """

    layer_cache: Dict[str, Layer] = {}
    tensor_cache: Dict[int, tf.Tensor] = {}
    model_input_tensors: List[tf.Tensor] = []
    output_tensors: List[tf.Tensor] = []
    for out_tensor in model.outputs:
        expanded_tensor = _expand_tensor(
            out_tensor, tensor_cache, layer_cache, model_input_tensors, layer_replacement_dict)
        output_tensors.append(expanded_tensor)
    converted_model = tf.keras.Model(inputs=model_input_tensors, outputs=output_tensors)
    return converted_model


def _expand_tensor(tensor: tf.Tensor, tensor_cache: Dict[int, tf.Tensor], layer_cache: Dict[str, Layer],
                   model_input_tensors: List[tf.Tensor], layer_replacement_dict: Dict[str, Layer]) -> tf.Tensor:
    tensor_id = id(tensor)
    expanded_tensor = tensor_cache.get(tensor_id)
    if expanded_tensor is not None:
        return expanded_tensor

    current_node = _get_node_from_tensor(tensor)
    current_layer = _get_layer_from_tensor(tensor)

    # Creating first model input layer
    if isinstance(current_layer, InputLayer):
        input_tensor = _create_input_tensor(tensor)
        tensor_cache[tensor_id] = input_tensor
        model_input_tensors.append(input_tensor)
        return input_tensor

    # get all input tensors
    node_input_tensors = []
    parent_nodes = current_node.parent_nodes
    for parent_node in parent_nodes:
        output_tensor = parent_node.outputs
        node_input_tensor = _expand_tensor(
            output_tensor, tensor_cache, layer_cache, model_input_tensors, layer_replacement_dict)
        node_input_tensors.append(node_input_tensor)

    # squeeze node_input_tensors
    if len(node_input_tensors) == 1:
        node_input_tensors = node_input_tensors[0]

    if is_custom_layer(current_layer):
        expanded_tensor = current_layer.call(node_input_tensors)
    elif current_layer.name in layer_replacement_dict:
        expanded_tensor = layer_replacement_dict[current_layer.name](node_input_tensors)
    else:
        expanded_tensor = current_layer(node_input_tensors)

    tensor_cache[tensor_id] = expanded_tensor
    return expanded_tensor


def _get_layer_from_tensor(tensor: tf.Tensor) -> Layer:
    history = _get_keras_history(tensor)
    layer = history.layer
    return layer


def _get_keras_history(tensor: tf.Tensor) -> KerasHistory:
    # pylint: disable=protected-access
    return tensor._keras_history


def _get_node_from_tensor(tensor: tf.Tensor) -> Node:
    history = _get_keras_history(tensor)
    layer = history.layer
    node_index = history.node_index
    node = layer.inbound_nodes[node_index]
    return node


def _create_input_tensor(input_tensor: tf.Tensor) -> tf.Tensor:
    history = _get_keras_history(input_tensor)
    input_layer = history.layer
    inp_config = input_layer.get_config()
    new_input_tensor = Input(**inp_config)
    _logger.debug(f"Input created, name: {new_input_tensor.name}, shape: {new_input_tensor.shape}")
    return new_input_tensor



import logging
from typing import Tuple, Dict

import tensorflow as tf
from tensorflow.keras import Model, Input
from tensorflow.keras.layers import Layer
from tensorflow.python.keras.models import Functional  # pylint: disable=no-name-in-module


def is_custom_layers(model: Model) -> bool:
    return any(is_custom_layer(layer) for layer in model.layers)


def should_replace_layers(model: Model, layer_replacement_dict: Dict[str, Layer]) -> bool:
    if layer_replacement_dict:
        for layer in model.layers:
            if layer.name in layer_replacement_dict:
                return True
    return False


def is_custom_layer(layer: Layer) -> bool:
    return not layer.__module__.startswith("tensorflow") or layer.__class__.__name__ == 'Functional'


def convert_subclassed_to_functional(model: Model, input_tensor_shape: Tuple[int, ...]) -> Functional:
    if not isinstance(model, Functional):
        input_tensor = Input(shape=input_tensor_shape)
        output_tensors = model.call(input_tensor)
        model = tf.keras.Model(inputs=input_tensor, outputs=output_tensors)

    return model


def configure_logger(verbose):
    logging_level = logging.INFO
    if verbose:
        logging_level = logging.DEBUG

    # pylint: disable=no-member
    loggers = [logging.getLogger(name)
               for name in logging.root.manager.loggerDict if name.startswith("leap_model_rebuilder")]
    root_logger = logging.getLogger()
    loggers.append(root_logger)
    for logger in loggers:
        logger.setLevel(logging_level)

    logging.basicConfig(format='%(asctime)s %(name)s %(levelname)s %(message)s', level=logging_level,
                        datefmt='%Y-%m-%d %H:%M:%S')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['leap_model_rebuilder']

package_data = \
{'': ['*']}

install_requires = \
['tensorflow>=2.0.0,<3.0.0']

setup_kwargs = {
    'name': 'leap-model-rebuilder',
    'version': '0.1.2',
    'description': '',
    'long_description': '# leap-model-rebuilder\nKeras model rebuilder tool for model rebuilding with alterations\n',
    'author': 'dorhar',
    'author_email': 'doron.harnoy@tensorleap.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/tensorleap/leap-model-rebuilder',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
